/***************************************************************
 * File: EmailMessage.cpp
 * Author: Keith Schwarz (htiek@cs.stanford.edu)
 *
 * This file contains the implementation of the EmailMessage class.
 * It includes the code for loading MIME types, doing Base64 conversion,
 * and composing MIME messages.  If you're interested in learning how
 * email messages are structured, then you might enjoy reading this file.
 * However, it's fairly technical and you may go blind reading some of
 * the more technical parts.
 *
 * You should not need to modify this file.
 */

#include "EmailMessage.h"
#include <sstream>
#include <fstream>
#include <iterator>
#include <map>
#include <cctype>
#include <algorithm>
#include <string>
using namespace std;

/* Implementation of ConvertToLowerCase */
static string ConvertToLowerCase(string input)
{
	transform(input.begin(), input.end(), input.begin(), ::tolower);
	return input;
}

/* Table of MIME types. */
static map<string, string> gMimeTypeTable;

static void InitializeMimeTypeTable()
{
	/* Only initialize once! */
	static bool initialized = false;
	if (initialized) return;

	initialized = true;

	/* Open the file of types and read data out. */
	ifstream input("mime-types.txt");
	string extension, type;

	while (getline(input, extension) && getline(input, type))
		gMimeTypeTable[ConvertToLowerCase(extension)] = type;
}

/* Emails containing attachments are structured as single files that contain sections
 * delimited by a "boundary string" which indicates to the email client that a new piece
 * of data is about to begin.  We choose the boundary string cs106lboundarydelineator,
 * which is unlikely to occur in messages.
 */
const string kBoundaryString = "cs106lboundarydelineator";

/* Data stored on-disk is encoded as a series of bytes with values between
 * 0 and 255.  However, SMTP was developed when most data was stored as
 * seven-bit values that range from 0 to 127.  This means that transmitting
 * binary data over a network is extremely difficult, since we somehow need
 * to compress values in the range 0 - 255 into the range 0 - 127 without
 * losing any data.  One particularly ingenious way of doing this uses a
 * technique called Base64 encoding.  The idea is to break the input into
 * groups of three-byte values (which contain twenty-four bits), then to
 * break this up into four six-bit values.  Six bit values range from 0
 * to 63, and can be represented using the values between 0 - 127.  However,
 * this encoding process is a bit tricky, and to save you much time and
 * frustration I've implemented Base64 encoding for you.  The implementation
 * is shown here.
 */
static string Base64Encode(string input)
{
	/* The six-bit groups ranging from 0 - 63 are each mapped to a unique character
	 * that is either a letter, a digit, or a special character.  The following table
	 * maps from values between 0 - 63 and the encoding characters.
	 */
	static const char kEncodeCharacters[] = "ABCDEFGHIJKLMNOP"
		                                    "QRSTUVWXYZabcdef"
											"ghijklmnopqrstuv"
											"wxyz0123456789+/";

	/* Because the input is broken up into groups of three bytes, we need to ensure 
	 * that the input is aligned to a multiple of three bytes.  This code pads the
	 * input appropriately.
	 */
	const size_t numAdded = (3 - input.length() % 3) % 3;
	for (size_t i = 0; i < numAdded; ++i)
		input += char(0);

	/* Break the input into groups of three and encode each block. */
	string result;
	for (size_t i = 0; i < input.length(); i += 3)
	{
		/* We are given the input broken up into three bytes:
		 * 00000000 11111111 22222222
		 *
		 * We want to interpret it as
		 * 00000011 11112222 22333333
		 *
		 * This requires bitwise operators to move the bits around appropriately.
		 * See me after lecture or send me an email for a description of exactly
		 * what's going on here.
		 */
		const size_t block0 = (input[i] >> 2) & (0x3F);
		const size_t block1 = ((input[i] & 0x3) << 4) | ((input[i + 1] >> 4) & 0xF);
		const size_t block2 = ((input[i + 1] & 0xF) << 2) | ((input[i + 2] >> 6) & 0x3);
		const size_t block3 = input[i + 2] & 0x3F;

		/* Convert each six-bit value into a Base64 character and append it to the
		 * result string.
		 */
		result += kEncodeCharacters[block0];
		result += kEncodeCharacters[block1];
		result += kEncodeCharacters[block2];
		result += kEncodeCharacters[block3];
	}

	/* Because we might have padded the message, we somehow need to tell the Base64
	 * decoder how many bytes we added.  This is done by adding an equals sign (=)
	 * to the end of the message string for each padded byte.
	 */
	for (size_t i = 0; i < numAdded; ++i)
		result[result.length() - 1 - i] = '=';

	return result;
}

/* When adding an attachment to a file, MIME requires us to store what type of
 * file is being added using a special code called a MIME type.  MIME types are
 * represented by the form category/specific, where category is a broad category
 * of what the file is and specific is a special identifier for that category.
 * For example, plain text is text/plain, HTML is text/html, PNG images are
 * image/png, etc.
 *
 * The list of MIME types is stored in an auxiliary file called mime-types.txt.
 * This function queries the table of values loaded from that file for info
 * about the MIME type.
 */
static string GetMimeTypeForFilename(string filename)
{
	/* Look for the last dot in the filename.  If we can't find a dot, give up
	 * and use the default.
	 */
	const size_t lastDot = filename.rfind('.');
	if (lastDot == string::npos)
		return "application/octet-stream";
	
	/* Actually get the extension. */
	const string extension = filename.substr(lastDot);

	/* Look up the MIME type in the table. */
	InitializeMimeTypeTable();

	map<string, string>::const_iterator itr = gMimeTypeTable.find(ConvertToLowerCase(extension));
	if (itr == gMimeTypeTable.end()) return "application/octet-stream";

	return itr->second;
}

/* Given a filename, loads the file, Base64 encodes it, then breaks it up into reasonably-sized
 * lines.
 */
static bool GetAttachmentContents(string filename, string& result)
{
	/* Open the file and give up if we can't for some reason. */
	ifstream input(filename.c_str(), ios::binary);
	if (input.fail())
		return false;

	/* Get the raw file contents.  This uses the istreambuf_iterator type, which we'll
	 * talk about later in the quarter.  At a high level, this means "grab all of the
	 * contents in the file and insert them into the string."
	 */
	string rawData;
	rawData.insert(rawData.begin(), istreambuf_iterator<char>(input), istreambuf_iterator<char>());

	/* Base64-encode the data. */
	rawData = Base64Encode(rawData);

	/* Break the input into 256-char lines, since SMTP doesn't accept single lines of
	 * input that are bigger than 1000 characters.
	 */
	static const size_t kCharsPerLine = 256;

	result.clear();
	for (size_t i = 0; i < rawData.length(); ++i)
	{
		result.push_back(rawData[i]);
		if ((i + 1) % kCharsPerLine == 0)
			result.push_back('\n');
	}

	return true;
}

/* Adding a recipient stores that recipient and the appropriate type in an
 * internal list.
 */
void EmailMessage::addRecipient(string rcpt, RecipientType type)
{
	to.push_back(make_pair(rcpt, type));
}

/* To get the number of recipients, we simply ask how many people are in the stored
 * list.
 */
size_t EmailMessage::numRecipients()
{
	return to.size();
}

/* Returns the recipient at the proper index in the list. */
string EmailMessage::getRecipient(size_t index)
{
	return to[index].first;
}

/* We only have one sender, so just store the proper value. */
void EmailMessage::setSender(string sender)
{
	from = sender;
}

/* Returns who the sender is. */
string EmailMessage::getSender()
{
	return from;
}

/* Sets the message subject. */
void EmailMessage::setSubject(string subj)
{
	subject = subj;
}

/* Writes the message text directly as it is. */
void EmailMessage::setMessageText(string msg)
{
	message = msg;
}

/* Given something of the form
 *
 * /dir1/dir2/dir3/.../dirn/filename.ext
 *
 * returns
 *
 * filename.ext
 */
string GetFilename(string filename)
{
	/* Grab everything from the last / forward. */
	string::size_type lastSlash = filename.rfind('/');
	if (lastSlash == string::npos)
		return "unknown-file";

	/* Get the string and drop out bad characters. */
	string result = filename.substr(lastSlash + 1);

	return result;
}

/* Adding an attachment actually loads the file contents and computes the MIME type
 * for that attachment.
 */
bool EmailMessage::addAttachment(string filename)
{
	Attachment att;
		
	/* Try to get the contents; fail if we can't. */
	if (!GetAttachmentContents(filename, att.data))
		return false;

	/* Get the MIME type and store it. */
	att.mimeType = GetMimeTypeForFilename(filename);

	/* Get the filename and store it. */
	att.filename = GetFilename(filename);

	/* Append the attachment to the list of attachments. */
	attachments.push_back(att);
	return true;
}

/* Generate a MIME email message.  This is perhaps the trickiest part of the code, since there
 * are a lot of details to get right.  The assignment handout is probably your best line of
 * defense for understanding what's going on here.
 */
string EmailMessage::toString()
{
	/* We'll use a stringstream to hold the resulting message, since it's really
	 * easy to work with.
	 */
	stringstream result;

	/* Describe the MIME format we're using.  We're using MIME 1.0 with a multipart message. */
	result << "MIME-Version: 1.0" << endl;
	result << "Content-Type: multipart/mixed; boundary=\"" << kBoundaryString << '"' << endl;

	/* Now, insert the message headers.  This starts with the From field. */
	result << "From: " << from << endl;

	/* Add all of the recipients using a different header for each. */
	for (size_t i = 0; i < to.size(); ++i)
	{
		/* Based on the type of the recipient, include a different header. */
		switch(to[i].second)
		{
		case To:
			result << "To: " << to[i].first << endl;
			break;
		case Cc:
			result << "Cc: " << to[i].first << endl;
			break;
		case Bcc:
			/* Do nothing for Bcc... they shouldn't show up in the email! */
			break;
		}
	}

	/* Add the subject header. */
	result << "Subject: " << subject << endl;

	/* The partiton between the message header and the message body is a newline.  We
	 * insert one here.
	 */
	result << endl;

	/* Some mail clients don't get MIME (virtually all modern clients do).  This code
	 * puts in a message to those users explaining why they will have some trouble reading
	 * the message.
	 */
	result << "You need a mail client that supports multipart email messages to see this message." << endl;

	/* Insert the actual message, Base64 encoded to avoid edge cases. */
	result << "--" << kBoundaryString << endl;
	result << "Content-Type: text/plain;" << endl;
	result << "Content-Transfer-Encoding: base64" << endl;
	result << endl;
	result << Base64Encode(message) << endl;

	/* Add all the attachments */
	for (size_t i = 0; i < attachments.size(); ++i)
	{
		/* Start a new boundary block. */
		result << "--" << kBoundaryString << endl;
		result << "Content-Type: " << attachments[i].mimeType << endl;
		result << "Content-Transfer-Encoding: base64" << endl;
		result << "Content-Disposition: attachment; filename=\"" << attachments[i].filename << "\";" << endl;
		result << endl;
		result << attachments[i].data << endl;
	}

	/* End with the boundary. */
	result << "--" << kBoundaryString << "--" << endl;

	return result.str();
}

/* The implementation of clear() uses the temporary object syntax to
 * construct a temporary EmailMessage and set this object to a copy
 * of the temporary.  Since the temporary initializes everything to
 * empty, this clears all fields.
 */
void EmailMessage::clear()
{
	*this = EmailMessage();
}
