/**********************************************************
 * File: SendMailWidget.cpp
 * Authors: Jake Becker (jabecker@stanford.edu)
 *          Keith Schwarz (htiek@cs.stanford.edu)
 *
 * Implementation of the main window class for the SendMail
 * program.  You should not need to use or modify the contents
 * of this file.
 */

#include <QtGui>
#include <cassert>
#include <vector>
#include "mail.h"
#include "SendMailWidget.h"
using namespace std;

/** Module-Specific Functions **/

/* The recipients stored in the recipients list are of the form
 * To: Email@Address.com  This function extracts the recipient's name
 * and returns it.
 */
static QString ExtractRecipient(QString str)
{
	str.remove(0, str.indexOf(':') + 2);

	return str;
}

/* This function returns the type of a recipient (see the above function comments for
 * a clarification.)
 */
static EmailMessage::RecipientType ExtractRecipientType(QString msg)
{
	if (msg.indexOf("To: ") == 0)
		return EmailMessage::To;
	if (msg.indexOf("Cc: ") == 0)
		return EmailMessage::Cc;
	if (msg.indexOf("Bcc: ") == 0)
		return EmailMessage::Bcc;

	assert (0 && "Unknown recipient type?");
	return EmailMessage::To; // Suppress warnings.
}

/** Implementation of SendMailWidget **/

/* The SendMailWidget constructor wires up all of the buttons and controls
 * to their proper handler functions.
 */
SendMailWidget::SendMailWidget() : addressBookDialog(addressList)
{
	ui.setupUi( this );

	/* Make the address book list display the proper information. */
	ui.addressBookListView->setModel(&addressListModel);
	resetRecipients();

	/* Wire up the buttons to the handlers. */
	connect( ui.attachButton, SIGNAL(clicked()), this, SLOT(addAttachment()) );
	connect( ui.clearAttachmentsButton, SIGNAL(clicked()), this, SLOT(clearAttachments()) );
	connect( ui.manageAddressesButton, SIGNAL(clicked()), this, SLOT(manageAddresses()) );
	connect( ui.sendButton, SIGNAL(clicked()), this, SLOT(sendMessage()) );
	connect( ui.toButton, SIGNAL(clicked()), this, SLOT(addToRecipient()) );
	connect( ui.ccButton, SIGNAL(clicked()), this, SLOT(addCcRecipient()) );
	connect( ui.bccButton, SIGNAL(clicked()), this, SLOT(addBccRecipient()) );
	connect( ui.removeButton, SIGNAL(clicked()), this, SLOT(removeRecipient()) );
	connect( &addressBookDialog, SIGNAL(accepted()), this, SLOT(resetRecipients()) );
}

/* Handler function invoked when the "add attachment" button is clicked. */
void SendMailWidget::addAttachment()
{
        /* Display the file-chooser dialog and pick an existing file. */
	QString filename = QFileDialog::getOpenFileName();

	/* If we actually chose something, go display it! */
	if( !filename.isNull() )
	        ui.attachmentList->addItem(filename);
}

/* Resets the list of attachments. */
void SendMailWidget::clearAttachments()
{
	ui.attachmentList->clear();
}

/* Handler which invokes the "manage addresses" dialog box. */
void SendMailWidget::manageAddresses()
{
	addressBookDialog.exec();
}

/* Checks that all information is valid, then sends the message. */
void SendMailWidget::sendMessage()
{
	if(!isValidInfo())
	{
		 QMessageBox::warning( this, tr("CS106L SendMail"), tr("Please specify at least one recipient for this message.") );
		 return;
	}

	/* Construct a new EmailMessage and populate it with our data. */
	EmailMessage message;
	message.setSender( UserName() );
	message.setSubject( ui.subjectLineEdit->text().toStdString() );
	message.setMessageText( ui.messagePlainTextEdit->toPlainText().toStdString() );

	/* Add all of the recipients. */
	for(int i = 0; i < ui.recipientListWidget->count(); ++i)
	{
		QString recipient = ui.recipientListWidget->item(i)->text();
		message.addRecipient(ExtractRecipient(recipient).toStdString(), ExtractRecipientType(recipient) );
	}

	/* Add all of the attachments. */
	for(int i = 0; i < ui.attachmentList->count(); ++i)
	{
		QString attachment = ui.attachmentList->item(i)->text();
		message.addAttachment( attachment.toStdString() );
	}

	/* Send the message and report the result. */
	if( SendEmail(message) )
	{
		QMessageBox::information(this, tr("CS106L SendMail"), tr("Email sent successfully!") );
		reset();
	}
	else
	{
		QMessageBox::critical(this, tr("CS106L SendMail"), tr("Sending email failed.") );
	}
}

/* Resets all of the controls. */
void SendMailWidget::reset()
{
	ui.attachmentList->clear();
	ui.subjectLineEdit->clear();
	ui.messagePlainTextEdit->clear();
	resetRecipients();
}

/* Checks whether the user's input forms a valid email.  We do only
 * a rudimentary check here.
 */
bool SendMailWidget::isValidInfo()
{
	return ui.recipientListWidget->count() != 0;
}

/* Populate the internal list of addresses with your custom address book. */
void SendMailWidget::readAddressBookIntoStringList(QStringList& addressList)
{
	addressList.clear();
	vector<string> addressbook = LoadAddressBook();
	for(vector<string>::iterator it = addressbook.begin(); it != addressbook.end(); ++it)
		addressList.append(it->c_str());
}

/* Removes all recipients from the recipient list. */
void SendMailWidget::resetRecipients()
{
	readAddressBookIntoStringList(addressList);
	remainingAddressesList = addressList;
	addressListModel.setStringList(remainingAddressesList);
	ui.recipientListWidget->clear();
}

/* Adds a new recipient to the message in response to a button click. */
void SendMailWidget::addRecipient(EmailMessage::RecipientType type)
{
  /* If the user didn't select anything, don't add anything! */
	if(!ui.addressBookListView->currentIndex().isValid())
		return;

	/* Get the selected element and remove it from the address book list. */
	QString str = remainingAddressesList.takeAt( ui.addressBookListView->currentIndex().row() );
	addressListModel.setStringList(remainingAddressesList);

	/* Based on the type of the recipient (To, Cc, Bcc), update the string
	 * appropriately for display in the recipients list.
	 */
	switch (type)
	{
	case EmailMessage::To:
		str = "To: " + str;
		break;
	case EmailMessage::Cc:
		str = "Cc: " + str;
		break;
	case EmailMessage::Bcc:
		str = "Bcc: " + str;
		break;
	default:
                assert (0 && "Unknown recipient type?");
	}

	/* Update the recipients list. */
	ui.recipientListWidget->addItem(str);
}

/* These handlers connect the "To", "Cc", and "Bcc" buttons to the appropriate
 * code for updating the lists.
 */
void SendMailWidget::addToRecipient()
{
	addRecipient(EmailMessage::To);
}

void SendMailWidget::addCcRecipient()
{
	addRecipient(EmailMessage::Cc);
}

void SendMailWidget::addBccRecipient()
{
	addRecipient(EmailMessage::Bcc);
}

/* Removes a recipient from the recipient list in response to the user
 * clicking the "X" button.
 */
void SendMailWidget::removeRecipient()
{
	QListWidgetItem *selection = ui.recipientListWidget->currentItem();
	if(selection == NULL)
		return;

	QString recipient = ExtractRecipient(selection->text());
	delete selection;
	ui.recipientListWidget->update();
	remainingAddressesList.append(recipient);
	addressListModel.setStringList(remainingAddressesList);
}
